;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;; See the file LICENCE for licence information.
(in-package :cl-l10n)

(defparameter *ignore-categories*
  (list  "LC_CTYPE"  "LC_COLLATE"))

(defun locale (loc-name &key (use-cache t) (errorp t))
  (let ((name (aif (position #\. loc-name)
                   (subseq loc-name 0 it)
                   loc-name)))
    (acond ((and (not name) (not errorp)) nil)
           ((and use-cache (get-locale name)) it)
           ((probe-file (merge-pathnames *locale-path* name))
            (load-locale name))
           ((not errorp) (warn "Can't find locale ~A." name))
           (errorp (locale-error "Can't find locale ~A." name)))))

(defun load-locale (name)
  (let ((path (merge-pathnames *locale-path* name)))
    (format t "~&;; Loading locale from ~A.~%" path)
    (let ((locale (make-instance 'locale :name name))
          (*read-eval* nil)
          (*print-circle* nil))
      (with-open-file (stream path
                       :external-format #+(and sbcl sb-unicode) :latin1 
                                        #-(and sbcl sb-unicode) :default)
        (multiple-value-bind (escape comment) (munge-headers stream)
          (awhile (next-header stream)
            (awhen (make-category locale it (parse-category it stream
                                                            escape comment))
              (setf (get-category (category-name it) locale) it)))))
      (setf (get-locale name) locale))))

(defun load-all-locales (&optional (*locale-path* *locale-path*))
  (dolist (x (directory (merge-pathnames *locale-path* "*")))
    (when (pathname-name x)
      (with-simple-restart (continue "Ignore locale ~A." x)
        (handler-case (load-locale (pathname-name x))
          (locale-error (c) (warn "Unable to load locale ~A. ~%~A." x c)))))))

(defvar *category-loaders*
  '(("LC_IDENTIFICATION" . load-identification)
    ("LC_MONETARY" . load-category)
    ("LC_NUMERIC" . load-category)
    ("LC_TIME" . load-category)
    ("LC_MESSAGES" . load-category)
    ("LC_PAPER" . load-category)
    ("LC_TELEPHONE" . load-category)
    ("LC_MEASUREMENT" . load-category)
    ("LC_NAME" . load-category)
    ("LC_ADDRESS" . load-category)))

(defun get-loader (name)
  (cdr (assoc name *category-loaders* :test #'string=)))

(defun make-category (locale name vals)
  (awhen (get-loader name)
    (funcall it locale name vals)))

(defun load-category (locale name vals)
  (declare (ignore locale))
  (let ((cat (make-instance 'category :name name)))
    (typecase vals 
      (category vals)
      (t (dolist (x vals)
           (setf (get-cat-val (car x) cat) (cdr x)))
         cat))))

(defvar *id-vals* 
  '(("title" . title)
    ("source" . source)
    ("language" . language)
    ("territory" . territory)
    ("revision" . revision)
    ("date" . date)
    ("categories" . categories)))
 

(defun load-identification (locale name vals)
  (declare (ignore name))
  (dolist (x *id-vals*)
    (aif (cdr (assoc (car x) vals :test #'string=))
         (setf (slot-value locale (cdr x)) 
               (remove #\" it)))))

(defun line-comment-p (line comment)
  (or (string= line "")
      (and (> (length line) 0) ;; Ignore a comment line
           (char= (schar line 0) comment))))
      

(defun copy-category (cat line)
  (let ((from (trim (subseq line (position #\Space line))
                    (cons #\" *whitespace*))))
    (handler-case (let* ((locale (locale from)))
                    (or (get-category cat locale)
                        (locale-error "No category ~A in locale ~A." 
                                      cat from)))
      (error (c) (locale-error "Unable to copy Category ~A from ~A. ~A."
                               cat from c)))))

(defun parse-category (name stream escape comment)
  (let ((end (mkstr "END " name))
        (ret nil))
    (loop for line = (read-line stream nil stream)
          until (eq line stream) do
      (cond ((line-comment-p line comment))
            ((search end line) (return-from parse-category ret))
            ((search "END" line) 
             (locale-error "End of wrong block reached ~S. Expected ~S." 
                    line end))
            ((and (> (length line) 3) (search "copy" line :end2 4))
             (return-from parse-category 
               (copy-category name line)))
            (t (push (get-value line stream escape) ret))))))

(defun munge-headers (stream)
  (let ((escape #\\) (comment-char #\#))
    (loop for line = (read-line stream nil stream)
          for i from 1 do
          ;; HACK We assume that if the escape and comment
          ;; lines don't appear right away that they don't exist
          ;; This is to work around lispworks being unable
          ;; to unread a line of text character by character.
      (cond ((> i 3) (return nil))
            ((line-comment-p line comment-char))
            ((search "escape_char" line)
             (setf escape 
                   (schar (cdr (get-value line stream escape)) 0)))
            ((search "comment_char" line)
             (setf comment-char
                   (schar (cdr (get-value line stream escape)) 0)))))
    (values escape comment-char)))



(defun get-full-line (line stream escape)
  (let ((length (length line)))
    (if (char= (elt line (1- length)) escape)
        (let ((next-line (read-line stream nil stream)))
          (if (eq next-line stream)
              (locale-error "EOF Looking for next line of ~A." line)
              (get-full-line (concatenate 
                              'string 
                              (subseq line 0 (1- length))
                              (trim next-line))
                             stream
                             escape)))
        line)))


(defun real-character (char)
  (let ((int (parse-integer (trim char (list* #\U #\> #\< *whitespace*)) 
                            :radix 16)))
    (handler-case (code-char int)
      (type-error (c)
        (declare (ignore c))
        (locale-error "Cannot represent ~A as a character." int)))))

(defvar *regex* '(:sequence 
                  #\< 
                  (:greedy-repetition 0 nil 
                   (:inverted-char-class #\> #\<)
                   :everything)
                  #\>))

(defun old-real-value (val)
  (aif (all-matches-as-strings *regex* val)
       (map #-lispworks 'string #+lispworks 
            'lw:text-string #'real-character it)
       val))

;; KLUDGE
(defun real-value (val)
  (let ((val (old-real-value val)))
    (if (string= val "\"\"")
        ""
        val)))


(defun parse-value (val)
  (let ((all-vals (split '(:char-class #\;) val)))
    (if (singlep all-vals)
        (real-value (car all-vals))
        (mapcar #'real-value all-vals))))

(defun get-value (line stream escape)
  "Return a cons containing the key of line and its value. 
   Honors lines ending with ESCAPE"
  (let* ((line (get-full-line line stream escape))
         (first-space (position-if #'(lambda (x)
                                       (or* (char= x #\Space #\Tab))) 
                                   line)))
    (if (null first-space)
        (locale-error "No Space in line ~A." line)
        (cons (trim (subseq line 0 first-space))
              (parse-value (trim (subseq line first-space)))))))

(defun next-header (stream)
  (loop for line = (read-line stream nil stream)
        until (eq line stream) do
    (if (and (> (length line) 3) (search "LC_" line :end2 3)
             (not (some #'(lambda (x)
                            (search x line :test #'string=))
                        *ignore-categories*))) 
        (return-from next-header line))))

(defun load-default-locale ()
  (setf *locale* (get-default-locale)))

(defun get-default-locale () 
  (or (locale (getenv "CL_LOCALE") :errorp nil)
      (locale (getenv "LC_CTYPE") :errorp nil)
      (locale "POSIX")))

(load-default-locale)


;; EOF